__doc__ = """
This module contains a class and a couple of functions.
The Konfabulator class handles loading up a config file (which must be in YAML format)
and checks for basic structure validity.
The class also allows a caller to specify the loading of specific connection classes
for use in the DBConnector.
"""

#################################################################################################

import sys, os
import re
import shutil
#from pprint import pprint

import yaml

from dbeif.utils.eif_exception import ConfigurationError, NonFatalConfigurationError
from dbeif.utils.security_manager import SecurityManager
from dbeif.utils.fuzzer import Fuzzer


#################################################################################################

class Konfabulator:
    """
        An instance of this class provides a means to read a configuration file
        in YAML format and separate the various sections out for easy consumption by
        holders of a Konfabulator instance.
        An instance also offers the capability to "fuzz" clear-text passwords in a
        configs file into an encoded (but not encrypted) form and to be able to handle
        defuzzing those encoded passwords for holders of the instance when they access
        those values via a dict like access.
    """

    def __init__(self, config_file_name, logger, cleartext_flag):
        logger.debug("initialization started...")
        #print("Konfabulator initialization started...")
        sys.stdout.flush()
        config_dir = 'configs'
        self.config_file_name = config_file_name
        self.log = logger
        self.cleartext_flag = cleartext_flag
        self.top_level_sequence = []
        self.config_file_path =  '%s/%s' %(config_dir, self.config_file_name)
        self.security_level = 'Encrypt'

        # check for existence, file-ness, readability of config_file_name
        if not os.path.exists(self.config_file_path):
            raise ConfigurationError('config file: %s not found' % self.config_file_path)
        if not os.path.isfile(self.config_file_path):
            raise ConfigurationError('config file: %s is not a file' % self.config_file_path)
        if not os.access(self.config_file_path, os.F_OK | os.R_OK):
            raise ConfigurationError('config file: %s not a readable file' % self.config_file_path)
        self.log.debug("config target item identified as a viable filesystem element...")

        try:
            cf = open(self.config_file_path, 'r', encoding='utf-8')
        except IOError as msg:
            raise ConfigurationError('Unable to open %s for reading, %s' % (self.config_file_path, msg))
        self.log.debug("config file opened for reading...")
        try:
            self.content = cf.read()
        except Exception as msg:
            raise ConfigurationError("ERROR:  unable to read content of file: %s, %s" % (self.config_file_path, msg))
        self.log.debug("config file contents have been read...")

        try:
            cf.close()
        except Exception as msg:
            raise ConfigurationError("ERROR:  unable to  close file: %s, %s" % (self.config_file_path, msg))
        self.log.debug("config file closed after reading...")

        basic_sanity, problem = self._checkConfigFileContentSanity()
        if not basic_sanity:
            raise ConfigurationError('Config file (%s) syntax/structure is incorrect, %s' % (self.config_file_path, problem),
                                     logger=logger)
        self.log.debug("config file contents pass the basic sanity checking...")

        complete_config = {}
        try:
            complete_config = yaml.load(self.content)
        except Exception as msg:
            raise ConfigurationError('Unable to parse %s successfully, %s' % (self.config_file_path, msg), logger=logger)
        self.log.debug("config file contents successfully interpreted as YAML structure...")

        top_key = None
        try:
            top_key = list(complete_config.keys())[0]
            self.config = complete_config[top_key]
        except Exception as msg:
            raise ConfigurationError('Unable to identify or use the top-most key  in the config from this list: %s' % ", ".join(list(complete_config.keys())))
        self.log.debug("identified the top key in the config as: '%s'" % top_key)

        conf_lines = []
        section_headers = []
        connector_header = []
        try:
            conf_lines = [line for line in self.content.split('\n') if line and not re.search(r'^\s*#', line)]
        except Exception as msg:
            raise ConfigurationError('Unable to extract conf_lines from the config file content, %s' % msg)
        self.log.debug("obtained config file contents as a list of %d strings..." % len(conf_lines))

        try:
            connector_header = [line for line in conf_lines if re.search(r'^[A-Z][A-Za-z_]+.+\s*:', line)]
        except Exception as msg:
            raise ConfigurationError('Unable to identify a connector_header in the config file content lines, %s' % msg)
        self.log.debug("extracted connector_header of '%s' from config file content lines..." % repr(connector_header))

        try:
            section_headers  = [line for line in conf_lines if re.search(r'^    [A-Z][A-Za-z_]+.*\s*:', line)]
        except Exception as msg:
            raise ConfigurationError('Unable to identify section_headers in the config file content lines, %s' % msg)
        self.log.debug("extracted these section_headers: %s" % ", ".join(section_headers))

        ##
        ##print("internal representation of the config file:")
        ##pprint(self.config, depth=4)
        ##print("---------------------------------------")
        ##

        # set up defaults for the Service section if that section isn't in the configs file
        if 'Service' not in self.config:
            self.config['Service'] = {}
            self.config['Service']['LogLevel'] = 'Info'
            self.config['Service']['Preview'] = False
            self.config['Service']['SecurityLevel'] = self.security_level
            section_headers.append('    Service:')
        if not 'LogLevel' in self.config['Service']:
            self.config['Service']['LogLevel'] = 'Info'
        if not 'Preview' in self.config['Service']:
            self.config['Service']['Preview'] = False
        if 'SecurityLevel' not in self.config['Service']:
            self.config['Service']['SecurityLevel'] = self.security_level
        else:
            if self.config['Service']['SecurityLevel'] in ['Cleartext', 'Encode', 'Encrypt']:
                self.security_level = self.config['Service']['SecurityLevel']
            else:
                problem = 'Invalid Security level: %s is used in the configuration' % self.config['Service'][
                    'SecurityLevel']
                raise ConfigurationError(problem)
        if self.cleartext_flag:
            self.security_level = 'Cleartext'
        self.log.debug("Service section of the config contains: %s" % repr(self.config['Service']))

        # set up empty default DB section if that section isn't in the configs file
        if 'DB' not in self.config:
            self.config['DB'] = {}
            section_headers.append('    DB:')

        if len(section_headers) < 2:
            raise ConfigurationError('Insufficient content in configs file: %s' % self.config_file_path)
        if len(section_headers) > 3:
            raise NonFatalConfigurationError('Excess or unrecognized content in configs file: %s' % self.config_file_path)
        self.log.debug("config file has the correct number of section headers")

        self.agicen_header = section_headers.pop(0).strip().replace(':', '')
        if not self.agicen_header.startswith('AgileCentral'):
            raise ConfigurationError('First section in configs file must be AgileCentral section')
        self.top_level_sequence.append('AgileCentral')
        self.log.debug("processed the AgileCentral section of the config file...")

        self.db_header = section_headers.pop(0).strip().replace(':', '')

        if self.db_header not in ['DB', 'Service']:
            problem = 'Second section in config file must be a DB or Service header'
            raise ConfigurationError(problem)
        self.top_level_sequence.append(self.db_header)
        self.log.debug("processed the DB section of the config file...")

        # because the EnvironmentKey needs an AgileCentral Workspace and Project, set that up here
        workspace = self.config['AgileCentral']['Workspaces'][0]['Workspace']
        self.config['AgileCentral']['Workspace'] = workspace
        self.config['AgileCentral']['Project'] = workspace[::-1]

        while section_headers:
            header = section_headers.pop(0).replace(':', '').strip()
            try:
                if header not in ['Service']:
                    problem = 'configs section header "%s" not recognized, ignored...' % header
                    raise ConfigurationError(problem)
                else:
                    self.top_level_sequence.append(header)
            except ConfigurationError as msg:
                pass
        self.log.debug("processed the section headers...")

        if 'Service' not in self.top_level_sequence:
            self.top_level_sequence.append('Service')
        if 'DB' not in self.top_level_sequence:
            self.top_level_sequence.append('DB')
        self.log.debug("initialization of the instance using config file contents completed successfully...")
        #print("Konfabulator initialization completed...")

    def applySecurityPolicy(self):
        if self.security_level.upper() != 'Cleartext'.upper():
            self.sec_mgr = SecurityManager(self, self.security_level, self.log)
            target_sections = [{'AgileCentral': ['Username', 'Password', 'APIKey', 'ProxyUsername', 'ProxyPassword']},
                               {'DB'          : ['User',     'Password']}]

            for section in  target_sections:
                self.sec_mgr.applyPolicyToSection(section)

        # now, get rid of the artifically added
        del self.config['AgileCentral']['Workspace']
        del self.config['AgileCentral']['Project']


    def _checkConfigFileContentSanity(self):
        sanity = True
        problem = ''
        has_tabs = [char for char in self.content if char == "\t"]
        if has_tabs:
            sanity = False
            problem = "Your config file contains tab characters which are not allowed in a YML file."
            return sanity, problem
        self.log.debug("No tab characters detected in the config file content")

        config_lines = self.content.split("\n")
        self.log.debug("config file content split into %d lines" % len(config_lines))
        yaml_lines = [line for line in config_lines if line and not re.search(r'^\s*#', line)]
        yaml_lines = [line for line in yaml_lines if not re.search(r'^---|\.\.\.$', line)]
        self.log.debug("internal strip of comment/extraneous lines from config file contents completed leaving %d config lines" % len(yaml_lines))
        yee = r' (include|exclude|Project)\s*:'  # yee --> YAML entry exclusions
        yaml_major_lines = [line for line in yaml_lines if not re.search(yee, line)]
        line_indents = [re.search(r'^(?P<indent>\s*)', line).group('indent') for line in yaml_major_lines]
        indent_levels = sorted(list(set([len(indent) for indent in line_indents])))
        outer_start = indent_levels[0]
        if outer_start:
            first_item = 0
        else:
            first_item = 1
        self.log.debug("found the first item in the config file as index: %d" % first_item)

        indent = indent_levels[first_item]
        line_indent = 4
        # for line_indent in indent_levels[first_item + 1:]:
        #     if line_indent % indent:
        #         # check to see if the "violation" involves one of 'exclude', 'include', 'AgileCentral_Project'
        #         #   parent ident = line_indent - 2
        #         #   if parent_ident is not  a violator  parent_ident % indent == 0, then skip calling this a problem...
        #         sanity = False
        #         problem = 'The file does not contain consistent indentation for the sections and section contents. '
        #         break

        if not sanity:
            self.log.warn("structure and/or content anomaly detected in the config file")
            first_offender_indent = line_indent
            for ix, config_line in enumerate(config_lines):
                mo = re.search("^(?P<spaces>\s+)(?P<non_space>\S)", config_line)
                if mo and mo.group('non_space') == '#':
                    continue
                if mo and len(mo.group('spaces')) == first_offender_indent:
                    offending_line_index = ix + 1
                    problem = problem + "The first occurrence of the problem is on line: %d" % offending_line_index
                    break

        return sanity, problem

    def topLevels(self):
        return self.top_level_sequence

    def topLevel(self, section_name):
        if section_name in self.top_level_sequence: #and section_name in self.config:
            return self.config[section_name]
        else:
            problem = 'Attempt to retrieve non-existent top level configs section for %s'
            raise ConfigurationError(problem % section_name)

    def connectionClassName(self, section_name, prefix=None):
        if section_name not in self.config:
            raise ConfigurationError(
                'Attempt to identify connection class name for %s, operation not supported' % section_name)
        if section_name not in ['AgileCentral', 'DB']:
            raise ConfigurationError('Candidate connection class name "%s" not viable for operation' % section_name)
        section = self.config[section_name]
        if section_name == 'AgileCentral':
            class_name = 'AgileCentralConnection'
        else:
            if section_name != 'DB':
                raise ConfigurationError('Unrecognized section naming a Connection subclass: "%s"' % section_name)
            class_name = "%sDBConnection" %  prefix

        return class_name